import datetime
import traceback

from django.db import transaction

from aio_client.base import RequestTypeEnum
from aio_client.provider.api import push_request
from aio_client.provider.models import PostProviderRequest
from django.conf import settings
from educommon.ws_log.models import SmevLog
from educommon.ws_log.models import SmevSourceEnum

from concentrator.settings import SMEV3_FORM_DATA_MESSAGE_TYPE
from concentrator.smev3_v321.base.utils import is_cancel_allowed
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.constants import DECLARATION_AWAITING_CODE
from concentrator.smev3_v321.constants import DECLARATION_AWAITING_COMMENT
from concentrator.smev3_v321.constants import DECLARATION_REVIEWED_CODE
from concentrator.smev3_v321.constants import DECLARATION_REVIEWED_COMMENT
from concentrator.smev3_v321.constants import DECLARATION_STATUS_MAP
from concentrator.smev3_v321.constants import DEFAULT_CODE_COMMENT
from concentrator.smev3_v321.constants import DIRECT_STATUS_MAP
from concentrator.smev3_v321.constants import DIRECTED_WITHOUT_RESPONSE
from concentrator.smev3_v321.constants import (
    NO_REQUEST_REQUIRED_DECLARATION_STATUSES)
from concentrator.smev3_v321.models import ApplicantAnswer
from concentrator.smev3_v321.models import DeclarationOriginMessageID
from concentrator.smev3_v321.order.constants import CreateOrderStatusMapper
from concentrator.smev3_v321.service_types import kinder_conc
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationStatusLog
from kinder.core.declaration_status.enum import DSS
from kinder.core.declaration_status.models import DeclarationStatusTransfer
from kinder.core.direct.models import DirectRefusalReason
from kinder.core.direct.models import DirectStatusLog
from kinder.core.direct.models import DRS
from kinder.core.direct.models import RefusalReasonInitiatorType
from kinder.core.direct.models import TEXT_CHANGE_STATUS
from kinder.core.group.models import Pupil


def get_code_and_comment(instance):
    """
    Возвращает код и комментарий для передачи сообщения при смене статуса
    заявления или направления

    :param instance: Заявление или направление
    :type instance: Union[Declaration, Direct]
    :return: Код и комментарий
    :rtype: Tuple[int, str]
    """
    
    if isinstance(instance, Declaration):
        code, comment = DECLARATION_STATUS_MAP.get(
            instance.status.code, DEFAULT_CODE_COMMENT)
        decl_status_log = DeclarationStatusLog.objects.filter(
            declaration=instance, status=instance.status
        ).latest('datetime')
        status_transfer = DeclarationStatusTransfer.objects.filter(
            from_status=decl_status_log.old_status,
            to_status=decl_status_log.status
        ).first()
        if status_transfer and status_transfer.auto_change_days:
            date_before = datetime.date.today() + datetime.timedelta(
                days=status_transfer.auto_change_days)
            doc_date_before = (
                f' в срок до {date_before.strftime(settings.DATE_FORMAT)} ')
        else:
            doc_date_before = ''

        if instance.status.code in DSS.status_queue_full():
            declaration_reviewed_check(instance)
            declaration_awaiting_direction_check(instance)
            return 0, ''

        data = {
            'mo': instance.mo.name,
            'comment': decl_status_log.comment,
            'mo_address': instance.mo.address_full,
            'doc_date_before': doc_date_before,
        }
    else:
        status_log = DirectStatusLog.objects.filter(
            direct=instance, status=instance.status
        ).latest('created_at')
        if instance.group.unit.days_for_direct_dogovor:
            dogovor_days = status_log.created_at + datetime.timedelta(
                days=instance.group.unit.days_for_direct_dogovor)
        else:
            dogovor_days = None
        data = {
            'doo': instance.group.unit.name,
            'group': instance.group.name,
            'age_cat': instance.group.age_cat.name,
            'document_details': instance.document_details,
            'mo': instance.declaration.mo.name,
            'unit_address': instance.group.unit.address_full,
        }
        if instance.status.code == DRS.REFUSED:
            reason = DirectRefusalReason.objects.filter(
                direct_status_log=status_log,
                reason__initiator_type=RefusalReasonInitiatorType.PARENT
            ).first()
            if (reason and reason.reason.initiator_type ==
                    RefusalReasonInitiatorType.EMPLOYEE):
                data.update({'reason': reason.reason.name})
                code, comment = DIRECT_STATUS_MAP.get(
                    DRS.REJECT, DEFAULT_CODE_COMMENT)
            else:
                code, comment = DIRECT_STATUS_MAP.get(
                    DRS.NEED_CHANGE, DEFAULT_CODE_COMMENT)
        elif instance.status.code == DRS.REGISTER:
            applicant_answer = ApplicantAnswer.objects.filter(
                direct=instance).first()
            if applicant_answer:
                if applicant_answer.answer:
                    code, comment = DIRECT_STATUS_MAP.get(
                        DIRECTED_WITHOUT_RESPONSE)
                else:
                    code, comment = DIRECT_STATUS_MAP.get(DRS.NEED_CHANGE)
            else:
                code, comment = DIRECT_STATUS_MAP.get(DRS.REGISTER)
        else:
            if status_log.old_status.code == DRS.REGISTER:
                date_before = status_log.created_at + datetime.timedelta(
                    days=instance.group.unit.days_for_reject_direct)
                data.update(
                    {'reason':
                     ' окончания срока действия направления '
                     f'{date_before.strftime(settings.DATE_FORMAT)}'})
            elif status_log.old_status.code == DRS.DOGOVOR:
                if dogovor_days is not None:
                    reason = (' окончания срока действия направления '
                              f'{dogovor_days.strftime(settings.DATE_FORMAT)}')
                else:
                    reason = ''
                data.update({'reason': reason})
            if dogovor_days is not None:
                dogovor_days = (
                    f' до {dogovor_days.strftime(settings.DATE_FORMAT)}')
            else:
                dogovor_days = ''
            data.update(
                {'dogovor_days': dogovor_days})
            code, comment = DIRECT_STATUS_MAP.get(
                instance.status.code, DEFAULT_CODE_COMMENT)

        pupil = Pupil.objects.filter(
            grup=instance.group, children=instance.declaration.children
        ).first()

        if pupil:
            data.update({'date_in_order_to_doo': pupil.date_in_order_to_doo})

        if instance.group.unit.days_for_reject_direct:
            date_before = status_log.created_at + datetime.timedelta(
                days=instance.group.unit.days_for_reject_direct)
            date_before = f' до {date_before.strftime(settings.DATE_FORMAT)}'
        else:
            date_before = ''

        data.update({'date_before': date_before})

    return code, comment.format(**data)


def push_change_order_info_request(
        declaration, org_code, comment, message_id,
        replay_to, log_to_smevlog=True
):
    """Отправляет сообщение ChangeOrderInfo"""

    response_body, response, error_traceback, exception = None, None, None, None

    try:
        response_body = render_type2xml(
            kinder_conc.FormDataResponseType(
                changeOrderInfo=kinder_conc.changeOrderInfoType(
                    orderId=kinder_conc.orderIdType(int(declaration.client_id)),
                    statusCode=kinder_conc.statusCodeType(
                        orgCode=org_code),
                    comment=comment,
                    cancelAllowed=is_cancel_allowed(declaration.status)
                )),
            name_type='FormDataResponse')

        response = push_request(PostProviderRequest(
            origin_message_id=message_id,
            body=response_body,
            message_type=SMEV3_FORM_DATA_MESSAGE_TYPE,
            attachments=None,
            content_failure_code=None,
            content_failure_comment='',
            replay_to=replay_to,
            is_test_message=False))
    except Exception as e:
        exception = e
        error_traceback = traceback.format_exc()

    if log_to_smevlog:
        SmevLog.objects.create(
            service_address=RequestTypeEnum.get_url(
                RequestTypeEnum.PR_POST),
            method_name='changeOrderInfo',
            method_verbose_name=(
                'FormData (Взаимодействие с формой-концентратором по СМЭВ 3)'),
            direction=SmevLog.OUTGOING,
            interaction_type=SmevLog.IS_SMEV,
            source=SmevSourceEnum.CONCENTRATOR,
            request=response_body,
            response=response.text if response else None,
            result=error_traceback
        )

    return response, exception


def declaration_reviewed_check(declaration):
    """Проверка для статуса концентратора Заявление рассмотрено

    :param declaration: Заявление
    :type declaration: Declaration
    :return: Признак удовлетворения условиям статуса
    :rtype: bool
    """

    if CreateOrderStatusMapper(
            declaration, None
    ).check_is_application_reviewed(declaration.status.code):
        message_id = declaration.declarationoriginmessageid.message_id
        comment = DECLARATION_REVIEWED_COMMENT.format(
            **{'order_id': declaration.client_id,
               'desired_date': declaration.desired_date.strftime(
                   settings.DATE_FORMAT)})
        push_change_order_info_request(
            declaration, DECLARATION_REVIEWED_CODE,
            comment, message_id,
            declaration.declarationoriginmessageid.replay_to)
        DeclarationOriginMessageID.objects.filter(
            declaration=declaration).update(reviewed_sent=True)
        return True

    return False


def declaration_awaiting_direction_check(declaration):
    """Проверка для статуса концентратора Ожидает направления

    :param declaration: Заявление
    :type declaration: Declaration
    :return: Признак удовлетворения условиям статуса
    :rtype: bool
    """

    if CreateOrderStatusMapper(
            declaration, None
    ).check_is_awaiting_direction(declaration.status.code):
        message_id = declaration.declarationoriginmessageid.message_id
        push_change_order_info_request(
            declaration, DECLARATION_AWAITING_CODE,
            DECLARATION_AWAITING_COMMENT, message_id,
            declaration.declarationoriginmessageid.replay_to)
        return True

    return False


def send_order_info_for_declaration(status_log):
    """
    Отправляет запрос changeOrderInfo для заявления по логу смены статуса

    :param status_log: Лог смены статуса
    :type status_log: DeclarationStatusLog
    """

    declaration_message_id = DeclarationOriginMessageID.objects.filter(
        declaration=status_log.declaration).first()

    if not status_log.old_status or not declaration_message_id or (
            status_log.old_status == status_log.status) or (
            status_log.status.code in NO_REQUEST_REQUIRED_DECLARATION_STATUSES
            and status_log.comment == TEXT_CHANGE_STATUS
    ):
        return

    message_id = declaration_message_id.message_id

    from concentrator.smev3_v321.base.tasks import (
        PushChangeOrderInfoRequestTask)
    transaction.on_commit(
        lambda: PushChangeOrderInfoRequestTask().apply_async(
            (status_log.declaration_id, None, None, message_id,
             declaration_message_id.replay_to,
             PushChangeOrderInfoRequestTask.DECLARATION_TYPE,
             status_log.declaration_id)
        )
    )
